@@ -29,6 +29,8 @@ Modified
 #include "vmbif.h"
 #include "vmfile.h"
 #include "vmlst.h"
+#include "vmstr.h"
+#include "vmrun.h"
 
 
 /* ------------------------------------------------------------------------ */
@@ -73,7 +75,7 @@ void CVmObjStrComp::delete_ext(VMG0_)
         size_t i;
         
         /* delete each first-level mapping table */
-        for (i = 0 ; i < sizeof(ext->equiv)/sizeof(ext->equiv[0]) ; ++i)
+        for (i = 0 ; i < countof(ext->equiv) ; ++i)
         {
             /* if this table is present, delete it */
             if (ext->equiv[i] != 0)
@@ -106,7 +108,7 @@ int CVmObjStrComp::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                             vm_obj_id_t self, vm_obj_id_t *source_obj,
                             uint *argc)
 {
-    unsigned short func_idx;
+    uint func_idx;
 
     /* translate the property into a function vector index */
     func_idx = G_meta_table
@@ -179,10 +181,10 @@ public:
         copy_limit = *value_ch_cnt;
         
         /* read the header values */
-        *ref_ch = str_->read_uint2();
+        *ref_ch = (wchar_t)str_->read_uint2();
         copy_size = *value_ch_cnt = str_->read_byte();
-        *uc_result_flags = (unsigned long)str_->read_int4();
-        *lc_result_flags = (unsigned long)str_->read_int4();
+        *uc_result_flags = str_->read_uint4();
+        *lc_result_flags = str_->read_uint4();
 
         /* limit copying to the actual buffer size */
         if (copy_size > copy_limit)
@@ -190,7 +192,7 @@ public:
 
         /* read the values */
         for (i = 0 ; i < copy_size ; ++i)
-            *val_buf++ = str_->read_uint2();
+            *val_buf++ = (wchar_t)str_->read_uint2();
 
         /* skip any values from the input that we weren't able to store */
         for ( ; i < *value_ch_cnt ; ++i)
@@ -209,7 +211,7 @@ protected:
 class CVmObjStrCompMapReaderList: public CVmObjStrCompMapReader
 {
 public:
-    CVmObjStrCompMapReaderList(const char *lst)
+    CVmObjStrCompMapReaderList(const vm_val_t *lst)
     {
         /* remember the list */
         lst_ = lst;
@@ -225,7 +227,7 @@ public:
                               wchar_t *val_buf, size_t *value_ch_cnt)
     {
         vm_val_t val;
-        const char *sublst;
+        vm_val_t sublst;
         const char *refstr;
         const char *valstr;
         size_t copy_rem;
@@ -236,18 +238,18 @@ public:
         copy_rem = *value_ch_cnt;
 
         /* retrieve the next element of the list */
-        CVmObjList::index_list(vmg_ &val, lst_, idx_);
+        lst_->ll_index(vmg_ &sublst, idx_);
 
         /* get the value as a sublist */
-        if ((sublst = val.get_as_list(vmg0_)) == 0)
+        if (!sublst.is_listlike(vmg0_) || sublst.ll_length(vmg0_) < 0)
             err_throw(VMERR_BAD_TYPE_BIF);
 
         /* retrieve the reference character string (sublst's 1st element) */
-        CVmObjList::index_list(vmg_ &val, sublst, 1);
+        sublst.ll_index(vmg_ &val, 1);
         refstr = val.get_as_string(vmg0_);
 
         /* retrieve the value string (sublst's 2nd element) */
-        CVmObjList::index_list(vmg_ &val, sublst, 2);
+        sublst.ll_index(vmg_ &val, 2);
         valstr = val.get_as_string(vmg0_);
 
         /* make sure the reference and value strings are indeed strings */
@@ -281,13 +283,13 @@ public:
         }
 
         /* get the upper-case flags (sublst's 3rd element) */
-        CVmObjList::index_list(vmg_ &val, sublst, 3);
+        sublst.ll_index(vmg_ &val, 3);
         if (val.typ != VM_INT)
             err_throw(VMERR_BAD_TYPE_BIF);
         *uc_result_flags = val.val.intval;
 
         /* get the lower-case flags (sublst's 4th element) */
-        CVmObjList::index_list(vmg_ &val, sublst, 4);
+        sublst.ll_index(vmg_ &val, 4);
         if (val.typ != VM_INT)
             err_throw(VMERR_BAD_TYPE_BIF);
         *lc_result_flags = val.val.intval;
@@ -298,7 +300,7 @@ public:
     
 protected:
     /* my list data */
-    const char *lst_;
+    const vm_val_t *lst_;
 
     /* the list index of the next mapping to retrieve */
     size_t idx_;
@@ -330,7 +332,7 @@ vm_obj_id_t CVmObjStrComp::create_from_stack(
 {
     size_t trunc_len;
     int case_sensitive;
-    const char *lst;
+    const vm_val_t *lst;
     vm_obj_id_t id;
     CVmObjStrComp *obj;
     size_t equiv_cnt;
@@ -370,35 +372,34 @@ vm_obj_id_t CVmObjStrComp::create_from_stack(
     else
     {
         size_t i;
+        int ec;
         
         /* get the list value from the argument */
-        lst = G_stk->get(0)->get_as_list(vmg0_);
-        if (lst == 0)
+        lst = G_stk->get(0);
+        if (!lst->is_listlike(vmg0_)
+            || (ec = lst->ll_length(vmg0_)) < 0)
             err_throw(VMERR_BAD_TYPE_BIF);
 
-        /* the list contains one entry per equivalence mapping */
-        equiv_cnt = vmb_get_len(lst);
-
         /* run through the list and count the value string characters */
-        for (i = 1, total_chars = 0 ; i <= equiv_cnt ; ++i)
+        for (i = 1, total_chars = 0, equiv_cnt = ec ; i <= equiv_cnt ; ++i)
         {
+            vm_val_t sublst;
             vm_val_t val;
-            const char *sublst;
             const char *strp;
             utf8_ptr ustrp;
 
             /* get this mapping from the list */
-            CVmObjList::index_list(vmg_ &val, lst, i);
+            lst->ll_index(vmg_ &sublst, i);
 
             /* make sure it's a sublist */
-            if ((sublst = val.get_as_list(vmg0_)) == 0)
+            if (!sublst.is_listlike(vmg0_))
                 err_throw(VMERR_BAD_TYPE_BIF);
 
             /* 
              *   get the second element of the mapping sublist - this is the
              *   value string 
              */
-            CVmObjList::index_list(vmg_ &val, sublst, 2);
+            sublst.ll_index(vmg_ &val, 2);
             if ((strp = val.get_as_string(vmg0_)) == 0)
                 err_throw(VMERR_BAD_TYPE_BIF);
 
@@ -510,7 +511,7 @@ void CVmObjStrComp::alloc_ext(VMG_ size_t trunc_len, int case_sensitive,
      *   we have no equivalence mappings installed yet, so clear out the
      *   first tier of the mapping array 
      */
-    for (i = 0 ; i < sizeof(ext->equiv)/sizeof(ext->equiv[0]) ; ++i)
+    for (i = 0 ; i < countof(ext->equiv) ; ++i)
         ext->equiv[i] = 0;
 
     /* load the mappings */
@@ -573,7 +574,7 @@ void CVmObjStrComp::save_to_file(VMG_ class CVmFile *fp)
  */
 ulong CVmObjStrComp::write_to_stream(VMG_ CVmStream *str, ulong *bytes_avail)
 {
-    wchar_t ref_ch;
+    wchar_t ref_ch_base;
     vmobj_strcmp_ext *ext = get_ext();
     size_t i;
     vmobj_strcmp_equiv ***p;
@@ -603,14 +604,14 @@ ulong CVmObjStrComp::write_to_stream(VMG_ CVmStream *str, ulong *bytes_avail)
     }
 
     /* write out the serialization structure header */
-    str->write_int2(ext->trunc_len);
-    str->write_int2(ext->case_sensitive ? 0x0001 : 0x0000);
-    str->write_int2(equiv_cnt);
-    str->write_int2(total_value_ch);
+    str->write_uint2(ext->trunc_len);
+    str->write_uint2(ext->case_sensitive ? 0x0001 : 0x0000);
+    str->write_uint2(equiv_cnt);
+    str->write_uint2(total_value_ch);
 
     /* run through our equivalence table again and write the mappings */
-    for (ref_ch = 0, i = 0, p = ext->equiv ;
-         i < sizeof(ext->equiv)/sizeof(ext->equiv[0]) ; ++i, ++p, ++ref_ch)
+    for (ref_ch_base = 0, i = 0, p = ext->equiv ; i < countof(ext->equiv) ;
+         ++i, ++p, ref_ch_base += 256)
     {
         vmobj_strcmp_equiv **ep;
         size_t j;
@@ -629,17 +630,17 @@ ulong CVmObjStrComp::write_to_stream(VMG_ CVmStream *str, ulong *bytes_avail)
                 wchar_t *vp;
                 
                 /* write the fixed part of the mapping */
-                str->write_int2(ref_ch);
+                str->write_uint2(ref_ch_base + j);
                 str->write_byte((uchar)(*ep)->val_ch_cnt);
-                str->write_int4((*ep)->uc_result_flags);
-                str->write_int4((*ep)->lc_result_flags);
+                str->write_uint4((*ep)->uc_result_flags);
+                str->write_uint4((*ep)->lc_result_flags);
 
                 /* write the value mapping characters */
                 for (k = (*ep)->val_ch_cnt, vp = (*ep)->val_ch ; k != 0 ;
                      --k, ++vp)
                 {
                     /* write this character */
-                    str->write_int2(*vp);
+                    str->write_uint2(*vp);
                 }
             }
         }
@@ -661,7 +662,7 @@ void CVmObjStrComp::count_equiv_mappings(size_t *equiv_cnt,
 
     /* run through our table and count up the mappings */
     for (*total_value_ch = 0, *equiv_cnt = 0, i = 0, p = ext->equiv ;
-         i < sizeof(ext->equiv)/sizeof(ext->equiv[0]) ; ++i, ++p)
+         i < countof(ext->equiv) ; ++i, ++p)
     {
         vmobj_strcmp_equiv **ep;
         size_t j;
@@ -721,7 +722,7 @@ int CVmObjStrComp::getp_calc_hash(VMG_ vm_obj_id_t /*self*/,
         err_throw(VMERR_BAD_TYPE_BIF);
 
     /* calculate the hash value and return it */
-    retval->set_int(calc_hash(strp + VMB_LEN, vmb_get_len(strp)));
+    retval->set_int(calc_str_hash(strp + VMB_LEN, vmb_get_len(strp)));
 
     /* discard gc protection */
     G_stk->discard();
@@ -730,94 +731,141 @@ int CVmObjStrComp::getp_calc_hash(VMG_ vm_obj_id_t /*self*/,
     return TRUE;
 }
 
+/*
+ *   Hash adder 
+ */
+struct StrCompHashAdder
+{
+    StrCompHashAdder(size_t trunc_len)
+    {
+        this->trunc_len = trunc_len;
+        this->has_trunc = (trunc_len != 0);
+        hash = 0;
+    }
+
+    int add(wchar_t ch)
+    {
+        /* add the character to the hash */
+        hash += ch;
+        hash &= 0xFFFF;
+
+        /* if there's a truncation limit, count this against the limit */
+        if (has_trunc && --trunc_len == 0)
+        {
+            /* we've reached the limit */
+            return FALSE;
+        }
+        else
+        {
+            /* haven't reached the limit yet */
+            return TRUE;
+        }
+    }
+
+    int done() const { return has_trunc && trunc_len == 0; }
+
+    /* truncation limit */
+    size_t trunc_len;
+
+    /* is there a truncation limit? */
+    int has_trunc;
+
+    /* the hash code */
+    unsigned int hash;
+};
+
 /*
  *   Calculate a hash value 
  */
-unsigned int CVmObjStrComp::calc_hash(const char *strp, size_t len)
+unsigned int CVmObjStrComp::calc_str_hash(const char *strp, size_t len)
 {
+    /* get my extension */
     vmobj_strcmp_ext *ext = get_ext();
-    utf8_ptr p;
-    unsigned int hash;
-    size_t char_limit;
 
     /* set up to scan the string */
-    p.set((char *)strp);
+    utf8_ptr p((char *)strp);
 
-    /*
-     *   Limit the scan to our truncation length, because we can't
-     *   distinguish hash buckets beyond the truncation point (if we did, a
-     *   truncated string wouldn't hash into the same bucket as a longer
-     *   string it matches; but all matching strings are required to go into
-     *   the same bucket, so such a hash mismatch is not allowed).
+    /* 
+     *   Scan the string.  Limit the scan to our truncation length, counting
+     *   substitution expansions and case folding expansions, because we
+     *   don't want to distinguish hash buckets beyond the truncation point.
+     *   If we did, a truncated string wouldn't hash into the same bucket as
+     *   a longer string it matches; all matching strings are required to go
+     *   into the same bucket, so we can't have such a hash mismatch.
      */
-    char_limit = ext->trunc_len;
-
-    /* scan the string */
-    for (hash = 0 ; len != 0 ; p.inc(&len))
+    StrCompHashAdder hash(ext->trunc_len);
+    for ( ; len != 0 ; p.inc(&len))
     {
-        wchar_t ch;
-        vmobj_strcmp_equiv **t1;
-        vmobj_strcmp_equiv *eq;
-
         /* get the current character */
-        ch = p.getch();
-
+        wchar_t ch = p.getch();
+        
         /* check for a substitution mapping for this character */
+        vmobj_strcmp_equiv **t1, *eq;
         if ((t1 = ext->equiv[(ch >> 8) & 0xFF]) != 0
             && (eq = t1[ch & 0xFF]) != 0)
         {
-            wchar_t *vp;
-            size_t vlen;
-            
             /* 
              *   This character has a mapping, so add the contribution from
              *   the canonical form of the character, which is the value
              *   side of the mapping.  
              */
+            wchar_t *vp;
+            size_t vlen;
             for (vp = eq->val_ch, vlen = eq->val_ch_cnt ; vlen != 0 ;
                  ++vp, --vlen)
             {
                 /* get this character */
                 ch = *vp;
 
-                /* convert to lower case if we're insensitive to case */
-                if (!ext->case_sensitive)
-                    ch = t3_to_lower(ch);
-
-                /* add it to the hash code */
-                hash += ch;
-                hash &= 0xFFFF;
-
-                /* if we've reached the truncation limit, we're done */
-                if (char_limit == 1)
-                    return hash;
-                else if (char_limit != 0)
-                    --char_limit;
+                /* 
+                 *   use case folding if we're insensitive to case and the
+                 *   character has a case folding
+                 */
+                const wchar_t *f;
+                if (!ext->case_sensitive && (f = t3_to_fold(ch)) != 0)
+                {
+                    /* add the folded expansion to the hash */
+                    for (const wchar_t *f = t3_to_fold(ch) ; *f != 0 ; ++f)
+                    {
+                        if (!hash.add(*f))
+                            return hash.hash;
+                    }
+                }
+                else
+                {
+                    if (!hash.add(ch))
+                        return hash.hash;
+                }
             }
         }
         else
         {
             /* 
-             *   if we are not sensitive to case, always use the lower-case
-             *   representation of a character for its hash value 
+             *   if we're not sensitive to case, and there's a case folding
+             *   available, use the case-folded representation of a character
+             *   for its hash value 
              */
-            if (!ext->case_sensitive)
-                ch = t3_to_lower(ch);
-            
-            /* add the contribution from this character */
-            hash += ch;
-            hash &= 0xFFFF;
-
-            /* if we've reached the truncation limit, we're done */
-            if (char_limit == 1)
-                return hash;
-            else if (char_limit != 0)
-                --char_limit;
+            const wchar_t *f;
+            if (!ext->case_sensitive && (f = t3_to_fold(ch)) != 0)
+            {
+                /* add the folded expansion to the hash */
+                for (const wchar_t *f = t3_to_fold(ch) ; *f != 0 ; ++f)
+                {
+                    if (!hash.add(*f))
+                        return hash.hash;
+                }
+            }
+            else
+            {
+                /* exact case only - add the character as-is */
+                if (!hash.add(ch))
+                    return hash.hash;
+            }
         }
     }
 
     /* return the hash code */
-    return hash;
+    return hash.hash;
 }
 
 /* ------------------------------------------------------------------------ */
@@ -874,7 +922,6 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
     utf8_ptr refp;
     unsigned long ret;
     int fold_case = !(ext->case_sensitive);
-    size_t valcharlen;
 
     /* set up to scan the strings */
     valp.set((char *)valstr);
@@ -884,33 +931,25 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
     ret = 0;
 
     /* scan the strings */
-    for (valcharlen = 0 ; vallen != 0 && reflen != 0 ; refp.inc(&reflen))
+    while (vallen != 0 && reflen != 0)
     {
-        wchar_t valch;
-        wchar_t refch;
-        vmobj_strcmp_equiv **t1;
-        vmobj_strcmp_equiv *eq;
-
         /* get each character */
-        valch = valp.getch();
-        refch = refp.getch();
+        wchar_t valch = valp.getch();
+        wchar_t refch = refp.getch();
 
         /* check for an exact match first */
         if (refch == valch)
         {
-            /* it's an exact match - skip this input character */
+            /* it's an exact match - skip this character in each string */
             valp.inc(&vallen);
-            ++valcharlen;
+            refp.inc(&reflen);
             continue;
         }
 
         /* check for a case-folded match if we're insensitive to case */
-        if (fold_case && t3_to_lower(valch) == t3_to_lower(refch))
+        if (fold_case
+            && t3_compare_case_fold_min(valp, vallen, refp, reflen) == 0)
         {
-            /* it's a case-folded match - skip this input character */
-            valp.inc(&vallen);
-            ++valcharlen;
-
             /* note in the flags that we have differing cases in the match */
             ret |= RF_CASEFOLD;
 
@@ -919,12 +958,11 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
         }
 
         /* check for a reference equivalence mapping */
+        vmobj_strcmp_equiv **t1;
+        vmobj_strcmp_equiv *eq;
         if ((t1 = ext->equiv[(refch >> 8) & 0xFF]) != 0
             && (eq = t1[refch & 0xFF]) != 0)
         {
-            wchar_t *vp;
-            size_t vlen;
-            
             /* 
              *   In case we match, apply the appropriate flags added for the
              *   equivalence mapping, based on the case of the first value
@@ -932,12 +970,15 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
              *   return failure, so it won't matter that we messed with the
              *   flags.)  
              */
-            ret |= (t3_is_upper(valch) ? eq->uc_result_flags
-                                              : eq->lc_result_flags);
+            ret |= (t3_is_upper(valch)
+                    ? eq->uc_result_flags
+                    : eq->lc_result_flags);
 
             /* match each character from the mapping string */
+            const wchar_t *vp;
+            size_t vlen;
             for (vp = eq->val_ch, vlen = eq->val_ch_cnt ;
-                 vallen != 0 && vlen != 0 ; ++vp, --vlen)
+                 vallen != 0 && vlen != 0 ; )
             {
                 /* get this character */
                 refch = *vp;
@@ -945,19 +986,18 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
                 /* if we have an exact match, keep going */
                 if (refch == valch)
                 {
-                    /* matched - skip this character and keep going */
+                    /* matched - skip this character in each string */
                     valp.inc(&vallen);
-                    ++valcharlen;
+                    ++vp, --vlen;
+
+                    /* keep going */
                     continue;
                 }
 
                 /* check for a case-folded match if appropriate */
-                if (fold_case && t3_to_lower(valch) == t3_to_lower(refch))
+                if (fold_case
+                    && t3_compare_case_fold_min(valp, vallen, vp, vlen) == 0)
                 {
-                    /* matched - skip this input character */
-                    valp.inc(&vallen);
-                    ++valcharlen;
-
                     /* note the case-folded match and keep going */
                     ret |= RF_CASEFOLD;
                     continue;
@@ -968,10 +1008,19 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
             }
 
             /* 
-             *   if we make it here, we matched the equivalence mapping -
-             *   we've already skipped the input we matched, so simply keep
-             *   going 
+             *   if we didn't make it to the end of the equivalence string,
+             *   we must have run out of source before we matched the whole
+             *   string, so we don'to have a match 
              */
+            if (vlen != 0)
+                return 0;
+
+            /* 
+             *   If we make it here, we matched the equivalence mapping.
+             *   We've already skipped the input we matched, so skip the
+             *   reference character and keep going.  
+             */
+            refp.inc(&reflen);
             continue;
         }
 
@@ -1008,6 +1057,7 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
          *   truncation length, or we don't allow truncation, it's not a
          *   match. 
          */
+        size_t valcharlen = utf8_ptr::s_len(valstr, valp.getptr() - valstr);
         if (ext->trunc_len != 0 && valcharlen >= ext->trunc_len)
         {
             /* 
@@ -1026,3 +1076,85 @@ unsigned long CVmObjStrComp::match_strings(const char *valstr, size_t vallen,
     }
 }
 
+/* 
+ *   Check for an approximation match to a given character.  This checks the
+ *   given input string for a match to the approximation for a given
+ *   reference character.  Returns the number of characters in the match, or
+ *   zero if there's no match.  
+ */
+size_t CVmObjStrComp::match_chars(const wchar_t *valstr, size_t vallen,
+                                  wchar_t refch)
+{
+    vmobj_strcmp_ext *ext = get_ext();
+    int fold_case = !(ext->case_sensitive);
+    vmobj_strcmp_equiv **t1;
+    vmobj_strcmp_equiv *eq;
+    wchar_t valch = *valstr;
+    const wchar_t *start = valstr;
+
+    /* check for an exact match first */
+    if (refch == valch)
+        return 1;
+            
+    /* check for a case-folded match if we're insensitive to case */
+    if (fold_case)
+    {
+        /* try a minimum case-folded match to the one-character ref string */
+        size_t ovallen = vallen, reflen = 1;
+        const wchar_t *refstr = &refch;
+        if (t3_compare_case_fold_min(valstr, vallen, refstr, reflen) == 0)
+        {
+            /* got it - return the length we consumed from the value string */
+            return ovallen - vallen;
+        }
+    }
+
+    /* check for a reference equivalence mapping */
+    if ((t1 = ext->equiv[(refch >> 8) & 0xFF]) != 0
+        && (eq = t1[refch & 0xFF]) != 0)
+    {
+        /* match each character from the mapping string */
+        const wchar_t *vp;
+        size_t vlen;
+        for (vp = eq->val_ch, vlen = eq->val_ch_cnt ;
+             vallen != 0 && vlen != 0 ; )
+        {
+            /* get this mapping character */
+            refch = *vp;
+
+            /* if we have an exact match, keep going */
+            if (refch == valch)
+            {
+                /* matched exactly - skip this character and keep going */
+                ++valstr, --vallen;
+                ++vp, --vlen;
+                continue;
+            }
+
+            /* check for a case-folded match if appropriate */
+            if (fold_case
+                && t3_compare_case_fold_min(valstr, vallen, vp, vlen) == 0)
+            {
+                /* matched - keep going */
+                continue;
+            }
+
+            /* no match */
+            return 0;
+        }
+
+        /* we've matched the mapping - return the match length */
+        return valstr - start;
+    }
+
+    /* it doesn't match any of the possibilities */
+    return 0;
+}
+
+/*
+ *   Get the truncation length 
+ */
+size_t CVmObjStrComp::trunc_len() const
+{
+    return get_ext()->trunc_len;
+}
